 /* Tables*/

 .data_type {
        font-size: 0.8em;
        font-weight: normal;
    }

    .column_name {
        font-size: 1.2em;
    }

    .info_items {
        margin: 10px 0;
        font-size: 0.8em;
    }

    .optimus_table td {
        border: 0px;
    }


    .optimus_table thead {
        border-bottom: 1px solid black;
    }
    .optimus_table{
        font-size: 12px;
    }

    .optimus_table tbody{
        font-family: monospace;
        border-bottom: 1px solid #cccccc;
    }

    .optimus_table tbody div{
        min-height:14px
    }

    /* Profiler */
        .main{
        width:100%;
        overflow:auto;
        border-bottom:1px solid #eeeeee;
        padding: 10px 0;
    }
    .panel_profiler{
        margin-right:2%;
        float:left;
        padding-bottom:2%;
    }
    .panel_profiler {
        font-family:monospace;
    }
    .title_profiler{
        padding:20px;
        background-color: #eeeeee
    }
    .info{
        overflow: auto
    }
    .main td, main th{
        padding:0em
    }
    .panel_profiler td {
        padding:0.2em
    }
    .optimus_table th {
        font-family:sans-serif;
    }

    .info_items{
        font-family:sans-serif;
        font-size:10px;
    }
